/*
   mem.c -- PIKT memory routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/mem.c,v 1.1 1999/06/14 21:48:26 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "common.h"

/* malloc() the specified amount of memory; if unsuccessful, presumably
   because of virtual memory exhaustion, log a warning and keep trying;
   if successful, return a generic pointer to the newly allocated memory
 */
void *
mymalloc (size_t size, char *func)
{
	void *ptr;
	static char msg[64];	/* should be plenty */

	while (! (ptr = malloc(size))) {
		sprintf(msg, "in %s(), malloc() failure", func);
		logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, FALSE, msg);
		sleep(MALLOC_TIME_TO_SLEEP);
	}
	return ptr;
}

/* calloc() the specified amount of memory; if unsuccessful, presumably
   because of virtual memory exhaustion, log a warning and keep trying;
   if successful, return a generic pointer to the newly allocated memory
 */
void *
mycalloc (size_t nobj, size_t size, char *func)
{
	void *ptr;
	static char msg[64];	/* should be plenty */

	while (! (ptr = calloc(nobj, size))) {
		sprintf(msg, "in %s(), calloc() failure", func);
		logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, FALSE, msg);
	}
	return ptr;
}
